import json
import click
import webbrowser
import pprint
from tempfile import NamedTemporaryFile
from max_ard.order import Order
import sys
from collections import defaultdict
from max_ard.exceptions import BadARDRequest

@click.group()
def order():
    ''' Tools for ordering ARD products '''
    pass


@click.command()
@click.argument('order_id', required=True)
@click.option('--format', help='Output the order status in other formats: `raw` for the full response JSON')
def status(order_id, format):
    ''' Gets the status of an order given the ORDER_ID '''

    try:
        s = Order.from_id(order_id)
    except Exception as e:
        print(str(e))
        sys.exit()
    if format == 'raw':
        print(s.response.dict())
    else:
        print(s.state)


@click.command()
@click.argument('order_id', required=True)
@click.option('--format', help='Output the order in other formats: None available')
def describe(order_id, format):
    ''' Outputs order details '''
    try:
        s = Order.from_id(order_id)
    except Exception as e:
        print(str(e))
        sys.exit()
    pp = pprint.PrettyPrinter(indent=4)
    if not s.finished:
        print('Select has not finished running')
        print('*******************')
        print('Request details:')
        pp.pprint(s.request.to_payload())
    if format is None:
        print(f'getting Select ID {order_id}')
        print('*******************')
        print('Request details:')
        pp.pprint(s.request.to_payload())
        print('*******************')
        print('\n\n')

class NumericType(click.ParamType):
    name = "numeric"

    def convert(self, value, param, ctx):
        try:
            return float(value)
        except TypeError:
            self.fail(
                "expected string for int() conversion, got "
                f"{value!r} of type {type(value).__name__}",
                param,
                ctx,
            )
        except ValueError:
            self.fail(f"{value!r} is not a value that can be converted to a float", param, ctx)

NUM_TYPE = NumericType()


@click.command()
@click.option('--destination', required=True, help='Destination of the tiles: an S3 bucket and optional prefix')
@click.option('--select_id', help='Order the results of Select with the Select ID')
@click.option('--acquisition_id', 'acquisitions', multiple=True, help='Order a given acquisition by acquisition ID, can be provided multiple times.' + \
    'If provided with a Select ID will only generate tiles for the provided IDs.')
@click.option('--intersects', help='Only generate tiles that intesect this geometry. Accepts WKT only.')
@click.option('--role_arn', help='A trusted Role ARN for the writer to assume so it can write tiles.')
@click.option('--dry-run', type=bool, default=False, help='Submit the order for basic validation without processing the order')
@click.option('--bbox', nargs=4, type=NUM_TYPE, help='Like --intersects, but takes a bounding box in the form --bbox XMIN YMIN XMAX YMAX')
@click.option('--bba', type=bool, default=False, help='Use Bundle Block Adjustment to align images within this order. See docs for limitations and best practices for using BBA.')
@click.option('--add-email', 'emails', multiple=True, help='Add an email address to receive order notifications. Can be used multiple times to add more than one email address.')
@click.option('--add-sns', 'topics', multiple=True, help='Add an AWS SNS topic to send nofications to. Can be used multiple times to add more than one topic.')
def submit(**kwargs):
    ''' Submit an ARD Order '''

    emails = kwargs.pop('emails', [])
    topics = kwargs.pop('topics', [])
    s = Order(**kwargs)
    if not emails and topics:
        if input("No notifications supplied, do you still want to order?") != "y":
            exit()
    for email in emails:
        s.add_email_notification(email)
    for topic in topics:
        s.add_sns_notification(topic)
    print('Request payload:')
    print(s.request.to_payload())


    if input("Are you sure you want to order? (y/n)") != "y":
        exit()

    s.submit()
    if kwargs['dry_run']:
        print('Dry run successful')
    else:
        print(f'Order {s.order_id} has been submitted')
        print(f'Run `max-ard order status {s.order_id}` to check the status')
        print('When the status is "SUCCEEDED" you can: ')
        print(f'Run `max-ard order describe {s.order_id}` to see a basic overview of the results')
        print('Run `max-ard order` to see all the `order` commands')


        


order.add_command(status)
order.add_command(describe)
order.add_command(submit)
