import click
import os
import json
import boto3, botocore
from max_ard.ard_collection import ARDCollection


@click.group()
def storage():
    ''' Commands related to ARD data storage'''
    pass

@click.command()
@click.argument('location')
def init(location):
    ''' Creates an S3 bucket if LOCATION does not exist, then authorizes the ARD data writer by adding a PutObject statement to the bucket policy of LOCATION '''

    location = location.replace('s3://', '')
    s3_resource = boto3.resource('s3')
    # check if the bucket exists
    try:
        s3_resource.meta.client.head_bucket(Bucket=location)
        print(f'Bucket exists at {location}')
    except botocore.exceptions.ClientError:
        s3_resource.create_bucket(Bucket=location) 
        print(f'Created bucket "{location}"')
    s3_client = boto3.client('s3')
    authorized = False
    try:
        policy = json.loads(s3_client.get_bucket_policy(Bucket=location)['Policy'])
    except botocore.exceptions.ClientError:
        policy = {
            "Version": "2012-10-17",
            "Statement": []
        }
    for statement in policy['Statement']:
        if statement['Sid'] == 'ArdS3Access':
            authorized = True
            break
    if not authorized:
        statement =  {
            "Sid":"ArdS3Access",
            "Effect":"Allow",
            "Principal":{
                "AWS":"arn:aws:iam::334489843805:root"
            },
            "Action":"s3:PutObject",
            "Resource":[
                f"arn:aws:s3:::{location}/*",f"arn:aws:s3:::{location}"
                ]
            }
        policy['Statement'].append(statement)
        s3_client.put_bucket_policy(Bucket=location, Policy=json.dumps(policy))
        print('Added ARD writer policy to bucket')
    else:
        print('Bucket already authorized')


@click.command()
@click.argument('location')
def revoke(location):
    ''' Removes the bucket policy statement from LOCATION, preventing the ARD writer from writing to the bucket '''

    location = location.replace('s3://', '')

    s3_resource = boto3.resource('s3')
    # check if the bucket exists
    try:
        s3_resource.meta.client.head_bucket(Bucket=location)
    except botocore.exceptions.ClientError:
        print(f"Can't find bucket {location}")
        exit()
    s3_client = boto3.client('s3')
    try:
        policy = json.loads(s3_client.get_bucket_policy(Bucket=location)['Policy'])
    except botocore.exceptions.ClientError:
        print('Bucket has no policy statements - nothing to revoke')
        exit()
    statements = []
    for statement in policy['Statement']:
        if statement['Sid'] != 'ArdS3Access':
            statements.append(statement)
    if len(statements) < len(policy['Statement']):
        if len(statements) == 0:
            s3_client.delete_bucket_policy(Bucket=location)
        else:
            policy['Statement'] = statements
            s3_client.put_bucket_policy(Bucket=location, Policy=json.dumps(policy))
        print('Removed ARD writer policy from bucket')
    else:
        print('Could not find ARD writer policy in bucket')

@click.command(name='list')
def list_():
    stores = os.environ.get('MAXARD_STORAGE')
    active = os.environ.get('MAXARD_ACTIVE_STORAGE')

    if stores is None:
        print('No storage locations have been set up yet')
        exit()
    
    for count, store in enumerate(stores.split(',')):
        line = f'{count+1}: {store}'
        if store == active:
            line += ' [ACTIVE]'
        print(line) 
    
@click.command(name='set')
@click.argument('location', required=False)
def set_(location):
    stores = os.environ.get('MAXARD_STORAGE')
    active = os.environ.get('MAXARD_ACTIVE_STORAGE')

    breakpoint()
    if stores is None and location is None:
        print('No storage locations have been set up yet')
        exit()
    
    try:
        stores = stores.split(',')
    except AttributeError:
        stores = []

    if location is None:
        for count, store in enumerate(stores):
            line = f'{count+1}: {store}'
            if store == active:
                line += ' [ACTIVE]'
            print(line) 
            selection = click.prompt('Select location to activate', type=int)
            if selection > len(stores):
                print('Location number was a not a valid value')
                exit()
            location = stores[selection-1]

    #verify_location(location)
    os.environ['MAXARD_ACTIVE_STORAGE'] = location
    if location not in stores: 
        stores.append(location)
        os.environ['MAXARD_STORAGE'] = ','.join(stores)

@click.command()
@click.argument('location')
def unset(location):
    stores = os.environ.get('MAXARD_STORAGE')
    active = os.environ.get('MAXARD_ACTIVE_STORAGE')

    if stores is None and location is None:
        print('No storage locations have been set up yet')
        exit()

    if location not in stores: 
        print('Location not found in list of storage locations')
        exit()
    
    try:
        stores = stores.split(',')
    except AttributeError:
        stores = []

    if location is None:
        for count, store in enumerate(stores):
            line = f'{count+1}: {store}'
            if store == active:
                line += ' [ACTIVE]'
            print(line) 
            selection = click.prompt('Select location to remove', type=int)
            if selection > len(stores):
                print('Location number was a not a valid value')
                exit()
            location = stores[selection-1]

    if active == location:
        print(f'{location} is no longer the active storage location')
        del os.environ['MAXARD_ACTIVE_STORAGE']
    stores.remove(location)
    os.environ['MAXARD_ACTIVE_STORAGE'] = ','.join(stores)
    print(f'Removed {location}')


@click.command()
@click.argument('location')
@click.option('--format', help='Output format, none supported yet')
@click.option('--aoi', help='WKT geometry to filter tiles to')
@click.option('--acq_id', 'acq_id_in', multiple=True, help='Limit result to this Acquisition ID (can use more than once')
@click.option('--zone', 'zone_in', multiple=True, help='Limit tiles to those in this zone. Can be used multiple times. This option can speed up scan time of large collections.')
@click.option('--earliest_date', help='Return results from on or after this date (YYYY-MM-DD)')
@click.option('--latest_date', help='Return results from this date or earlier (YYYY-MM-DD')
def describe(**kwargs):
    ''' Gives a summary of ARD tiles in LOCATION (S3 or local path) '''

    format = kwargs.pop('format', None)   
    location = kwargs.pop('location', None) or os.environ.get('MAXARD_ACTIVE_STORAGE', None)
    assert location, 'You need to have an active storage location or provide one via --location <path>'
    c = ARDCollection(location, **kwargs)    
    if format is None:
        print(f'Examining ARD data in {location}...')
        # this will trigger access, see if it fails
        try:
            count = len(c.acq_ids)
        except PermissionError:
            print(f'Your credentials for accessing {location} are invalid or expired')
            exit()
        if count == 0:
            print('No ARD tiles found')
            exit()
        print(f'{count} acquisitions')
        print(f'{len(c.tiles)} tiles')
        print(f'Earliest date: {c.start_date}')
        print(f'Latest date: {c.end_date}')
        zones = ','.join([str(z) for z in c.zones])
        print(f'UTM zones covered: {zones}')
    else:
        print('No other formats available at this time')

def get_location(location):
    stores = os.environ.get('MAXARD_STORAGE')
    active = os.environ.get('MAXARD_ACTIVE_STORAGE')

    if stores is None:
        print('No storage locations have been set up yet')
        exit()

def get_filesystem(location):
    pass


storage.add_command(init)
storage.add_command(revoke)
#storage.add_command(list_)
#storage.add_command(set_)
storage.add_command(describe)