from functools import lru_cache
from typing import Union, List
from posixpath import join as urljoin
from collections import defaultdict
from pathlib import Path
import os
import re
import json
import operator
import warnings

import requests
from pydantic import BaseModel
from shapely.geometry import mapping, GeometryCollection
from max_ard.session import get_session
from max_ard.exceptions import BadARDRequest, NotFinished, NotSubmitted, ARDServerException
from maxar_canvas_grid import Cell
try:
    from functools import cached_property
except ImportError:
    from backports.cached_property import cached_property

from max_ard.session import ARD_HOSTNAME

GET_SELECT_URL = f'https://{ARD_HOSTNAME}/select/request' # /{id}
SELECT_FILES_URL = f'https://{ARD_HOSTNAME}/select/files' # /{id}.{ext}
POST_SELECT_URL = f'https://{ARD_HOSTNAME}/select'

##############
# Validation decorators
#############


def needs_submittal(func):
    def func_wrapper(instance, *args, **kwargs):
        if not instance.submitted:
            raise NotSubmitted(f"Can't access {func.__name__} until Select has been submitted")
        res = func(instance, *args, **kwargs)
        return res
    return func_wrapper


def needs_completion(func):
    def func_wrapper(instance, *args, **kwargs):
        if not instance.finished:
            raise NotFinished(f"Can't access {func.__name__} until Select has finished running")
        res = func(instance, *args, **kwargs)
        return res
    return func_wrapper


############
# Objects
###########


class SelectRequest(BaseModel):
    ''' An object to hold the request to the Select Service '''

    acq_ids: List[str] = None
    datetime: str = None
    stack_depth: int = None 
    intersects: str = None
    #bbox = Union(List, None) = None
    query: dict = {}

    def to_payload(self):
        ''' sends only parameters with values '''
        params = self.dict().copy()
        return {k:v for k,v in params.items() if v not in [None, [], ()]}

    class Config:
        extra = 'allow'
        validate_assignment = True



class SelectReponse(BaseModel):
    ''' an object to hold the Select API response '''

    id: str
    state: str
    response_timestamp = str
    request_details: dict = None
    usage: dict = None
    links: dict = None
    error: str = None
    message: str = None

    class Config:
        allow_mutation: False



class Select():
    ''' An ARD API Select object

        An instance stores both the request and the response'''

    # store an authenticated Requests session in the class
    api_session = get_session()

    def __init__(self,
        acq_ids = None,
        datetime = None,
        intersects = None,
        bbox = None,
        query = {},
        stack_depth=None,
        ):
    
        # if the intersects is a Shapely geom, convert it to wkt
        try:
            intersects = intersects.wkt
        except:
            pass

        self.request = SelectRequest(
            acq_ids=acq_ids,
            datetime=datetime,
            intersects=intersects,
            bbox=bbox,
            stack_depth=stack_depth,
            query=query
        )
        self.response = None
        self.submitted = False


    @property
    @needs_submittal
    def finished(self):
        ''' The Select is finished computing and results are available '''

        return self.state == 'FINISHED'


    @property
    @needs_submittal
    def state(self):
        ''' State of the select process: 'RUNNING', 'FINISHED', or 'FAILED' '''
        if self.response.state == 'RUNNING':
            response = self.get_select(self.select_id)
            self.response = SelectReponse(**response) 
        return self.response.state


    @property
    @needs_submittal
    def select_id(self):
        ''' ID of the Select '''

        return self.response.id


    @property
    @needs_submittal
    @needs_completion
    def usage(self):
        ''' Dictionary of data usage metrics '''

        return self.response.usage

        
    @classmethod
    def from_id(cls, select_id):
        ''' Create a Select object from an ID '''

        instance = cls()
        response = cls.get_select(select_id)
        instance.response = SelectReponse(**response) 
        if 'request_details' in response:
            instance.request = SelectRequest(**response['request_details'])
        instance.submitted = True
        return instance

    @classmethod
    def get_select(cls, select_id):
        ''' Gets data from the API about a select '''

        url = urljoin(GET_SELECT_URL, select_id)
        r = cls.api_session.get(url)
        if r.status_code == 400:
            raise BadARDRequest(r.json()['message'])
        # server error
        elif r.status_code != 200:
            raise ARDServerException(r.json()['message'])
        # success
        return r.json()


    @classmethod
    def send_select(cls, payload):
        ''' Send a select request to the API '''

        r = cls.api_session.post(POST_SELECT_URL, json=payload)
        # bad format
        if r.status_code == 400:
            raise BadARDRequest(r.json()['message'])
        # server error
        elif r.status_code != 200:
            raise ARDServerException(r.json()['message'])
        # success
        return r.json()
    

    def submit(self):
        ''' Submits a select request to the API '''

        response = self.send_select(self.request.to_payload())
        if self.submitted:
            warnings.warn('The Select has already been submitted')
            return
        self.submitted = True
        self.response = SelectReponse(**response)
        # update request in case defaults were applied
        self.request = SelectRequest(**self.response.request_details)
        
    @lru_cache()
    @needs_submittal
    @needs_completion
    def get_link_contents(self, name):
        temp_url = self.get_signed_link(name)
        # unauthenticated, don't send token
        r = requests.get(temp_url)
        r.raise_for_status()
        return r.text

    @needs_submittal
    @needs_completion
    def get_signed_link(self, name):
        url = self.response.links[name]
        r = self.api_session.get(url)
        r.raise_for_status()
        return r.json()['download_link']

    @needs_submittal
    @needs_completion
    def copy_file(self, name, dir='.'):
        #TODO get the output filename
        path = Path(dir, f'{self.select_id}.{name}')
        with open(path, 'w') as out:
            out.write(self.get_link_contents(name))

    @cached_property
    @needs_submittal
    @needs_completion
    def results(self):
        return SelectResult.from_geojson(json.loads(self.get_link_contents('geojson')))



class SelectResult():

    def __init__(self):
        self.acquisitions = defaultdict(dict)
        self.cells = []
        self.count = 0 

    @classmethod
    def from_geojson(cls, geojson):
        instance = cls()
        for tile in geojson['features']:
            instance.cells.append(tile['properties']['cell_id'])
            for feature in tile['properties']['best_matches']:
                instance.count += 1
                instance.acquisitions[feature['acquisition_id']] = {
                    'date': feature['date'] 
                }
        return instance

    @property
    def __geo_interface__(self):
        geoms = [Cell(cell).geom_WGS84 for cell in self.cells]
        return mapping(GeometryCollection(geoms))

    def __repr__(self):
        return f'<SelectResult ({self.count} tiles in {len(self.acqs.keys())} acquisitions) >'


